use ark_ff::PrimeField;
use ark_poly::DenseUVPolynomial;
use ark_std::ops::Div;
use rs_merkle::algorithms::Sha256;
use rs_merkle::Hasher;
use tracing::{debug, info};

use crate::{algebra, error::KomodoError, fec};
use fri::{
    frida::{FridaBuilder, FridaCommitment},
    rng::FriChallenger,
    utils::{to_evaluations, HasherExt, MerkleProof},
};

/// representation of a block of proven data.
///
/// this is a wrapper around a [`fec::Shard`] with some additional cryptographic
/// information that allows to prove the integrity of said shard.
#[derive(Clone, PartialEq)]
pub struct Block<F: PrimeField, H: Hasher> {
    pub shard: fec::Shard<F>,
    proof: MerkleProof<H>,
    commit: FridaCommitment<F, H>,
    position: usize,
}

pub fn encode<F: PrimeField>(bytes: &[u8], k: usize, n: usize) -> Vec<fec::Shard<F>> {
    debug!("splitting bytes into rows");
    let elements: Vec<F> = algebra::split_data_into_field_elements(bytes, k);
    let rows = elements.chunks(k).map(|c| c.to_vec()).collect::<Vec<_>>();
    info!(
        "data is composed of {} rows and {} elements",
        rows.len(),
        elements.len()
    );

    let hash = Sha256::hash(bytes).to_vec();

    let w = F::get_root_of_unity(n as u64).unwrap();

    rows.into_iter()
        .enumerate()
        .map(|(i, r)| fec::Shard {
            k: k as u32,
            linear_combination: (0..k).map(|j| w.pow([(i + j) as u64])).collect(),
            hash: hash.clone(),
            data: to_evaluations(r, n),
            size: bytes.len(),
        })
        .collect::<Vec<_>>()
}

pub fn prove<const N: usize, F: PrimeField, H: Hasher, P>(
    shards: Vec<fec::Shard<F>>,
    blowup_factor: usize,
    remainder_plus_one: usize,
    nb_queries: usize,
) -> Result<Vec<Block<F, H>>, KomodoError>
where
    P: DenseUVPolynomial<F>,
    for<'a, 'b> &'a P: Div<&'b P, Output = P>,
    <H as rs_merkle::Hasher>::Hash: AsRef<[u8]>,
{
    let builder = FridaBuilder::<F, H>::new::<N, _>(
        &shards.iter().map(|s| s.data.clone()).collect::<Vec<_>>(),
        FriChallenger::<H>::default(),
        blowup_factor,
        remainder_plus_one,
        nb_queries,
    );

    let commit = FridaCommitment::from(builder.clone());

    Ok(shards
        .iter()
        .enumerate()
        .map(|(i, s)| Block {
            shard: s.clone(),
            proof: builder.prove_shards(&[i]),
            commit: commit.clone(),
            position: i,
        })
        .collect())
}

pub fn verify<const N: usize, F: PrimeField, H: Hasher, P>(
    block: Block<F, H>,
    domain_size: usize,
    nb_queries: usize,
) -> Result<(), KomodoError>
where
    P: DenseUVPolynomial<F>,
    for<'a, 'b> &'a P: Div<&'b P, Output = P>,
    <H as rs_merkle::Hasher>::Hash: AsRef<[u8]>,
{
    block
        .commit
        .verify::<N, _>(
            FriChallenger::<H>::default(),
            nb_queries,
            block.shard.k as usize,
            domain_size,
        )
        .unwrap();

    assert!(block.proof.verify(
        block.commit.tree_root(),
        &[block.position],
        &[H::hash_item(&block.shard)],
        domain_size,
    ));

    Ok(())
}
